#!/usr/bin/perl
#
# Transcode Recording files to MP3 format
# so they can be streamed by streaming server
#

use strict;
use Data::Dumper;
use File::Find;
use File::Copy;
use File::Path;
use File::Basename;
use threads;
use threads::shared;
use Thread::Queue;
use POSIX qw/ceil/;
use IO::Handle;

my @ast_format = (
	"ulaw",
	"alaw",
	"gsm",
	"slin",
	"g722",
	"g729"
);

my %sox_format = (
	"ulaw" => "ul",
	"alaw" => "al"
);

#my %ffmpeg_format = (
#	"ulaw" => "mulaw",
#	"alaw" => "alaw"
#);

my %options;
my $single_file;
my $server_id;
my $file_queue = Thread::Queue->new();
my $file_total_size;

# We assume it takes the server 1 second to process 1 MBytes data
# The value gets dynamically calculated.
my $transcoding_seconds_per_mbytes = 1;

open(my $LOG, ">>/var/log/trans_recordings.log") or die("Unable to open /var/log/trans_recordings.log: $!\n");

#if ($ARGV[0] eq "-s") {
#    exit_trans_recordings();
#}

if(@ARGV == 2)
{
	($single_file, $server_id) = @ARGV;
}

read_conf() or die "Failed to read config file\n";

if(defined($single_file))
{
	print "Translate single file: $single_file\n";
	do_single_trans($single_file, $server_id);
	print "Done\n";
	exit;
}
else
{
#	my $pid = fork();
#	if(not defined $pid) {
#		die "Fork failed\n";
#	} elsif (0 == $pid) { # child
		open(my $fh, '>', "/var/run/trans_recordings.pid") or die $!;
		print $fh "$$";
		close($fh);

		print "Trancoding process [$$] started.\n";
		addlog("Trancoding process [$$] started.\n");	

		while(1) {
			my $num = do_batch_trans();
			if($num == 0) {
				sleep(30);
			}
		}
#	}
}

close $LOG;
exit;

sub addlog
{
        my $msg = shift;
	print $LOG "[".localtime()."]".$msg;
	$LOG->autoflush(1);
}

sub adddbg
{
	my $msg = shift;
	return if( ($options{'options'}{'debug'} eq 'no') || ($options{'options'}{'debug'} eq 'false') );
	print $LOG "[".localtime()."]".$msg;
}

sub exit_trans_recordings
{
	open(my $fh, '<', "/var/run/trans_recordings.pid") or die "Trans_recordings script is not running\n";

	my $pid = <$fh>;
	close($fh);

	# kill server
	kill 9, $pid;

	# clean up proccess
	my $exit_code = waitpid($pid, 0);
	print "Trans_recordings process terminated.\n";
	exit;
}


sub read_conf
{
	open(CONF, "/usr/local/rserver/trans_recordings.conf") or die "Couldn't open config file";

	my $key;
	while(my $line = <CONF>)
	{
		next if($line =~ /^;.*$/);

		if ($line =~ /\[(.*)\]/) {
			$key = $1;
			next;
		} else {
			$line =~ /(.*)=(.*)/;
			share($options{$key}{$1});
			$options{$key}{$1} = $2;
		}
	}

	close CONF;
	if (!defined($options{directories}{input})
		|| !defined($options{directories}{output})) {
		return 0;
	}
	
	addlog("[Directories]\n");
	addlog("".$options{directories}{input}." => ".$options{directories}{output}."\n");
	return 1;
}

# translate recording file under /var/fonality/recordall/recordings/<sid>/ and save the transcoded file to
# /var/fonality/recordall/tivo/<sid>/

sub do_single_trans
{
	my $single_file = shift;
	my($filename, $directories, $suffix) = fileparse($single_file, qr/\.[^.]*~$/);	

	print "Directory: $directories\n";
	print "Filename: $filename\n";
	print "Suffix: $suffix\n";

	my $ext;
	my $found = 0;
	foreach my $format (@ast_format)
	{
		if($suffix =~ /\.\Q$format\E~$/)
		{
                        $found = 1;
                        $ext = $format;
                        last;
		}
	}
	return undef unless($found);

	my $output_filename;
	my $in_or_out;
	if($filename =~ /(.*)\-in$/)
	{
		$output_filename = $1;
		$in_or_out = 1;
	} elsif($filename =~ /(.*)\-out$/) {
		$output_filename = $1;
		$in_or_out = 0;
	} else {
		print "Unknown file: ".$single_file;
		return undef;
	}

	# get the output directory
	my $output_dir = "/var/fonality/recordall/tivo/".$server_id."/";
	my $output_pathname = $output_dir.$output_filename;

	my $save_u = umask();
	umask(0);
	mkpath($output_dir, { verbose => 1, mode => 0777 }) unless(-e $output_dir);
	umask($save_u);

	unlink $output_pathname.".wav";

	# trancode all formats other than alaw & ulaw into ulaw
        my $in_orig_file;
        my $out_orig_file;
        my $command;
        if($in_or_out)
        {
		$in_orig_file = $single_file;
		$out_orig_file = $directories.$output_filename."-out.".$ext."~";
	} else {
		$out_orig_file = $single_file;
		$in_orig_file = $directories.$output_filename."-in.".$ext."~";
	}

	copy($in_orig_file, $output_pathname."-in.".$ext);
	copy($out_orig_file, $output_pathname."-out.".$ext);

	if($ext ne "ulaw" && $ext ne "alaw")
	{
		if(-e $in_orig_file)
		{
                        $command = "/usr/sbin/asterisk -rx \"file convert ".$output_pathname."-in.".$ext." ".$output_pathname."-in.ulaw\"";
                        run_cmd(0, $command) || return;
			unlink $output_pathname."-in.".$ext;
		}

        	if(-e $out_orig_file)
		{
        	        $command = "/usr/sbin/asterisk -rx \"file convert ".$output_pathname."-out.".$ext." ".$output_pathname."-out.ulaw\"";
			run_cmd(0, $command) || return;
			unlink $output_pathname."-out.".$ext;
		}

		$ext = "ulaw";
	}
	
	# merge in & out into one and encode it as mp3 format
	my $in_middle_file = $output_pathname."-in.".$ext;
	my $out_middle_file = $output_pathname."-out.".$ext;
	my $type = $sox_format{$ext};
	my $wav_options = "-t wav -r 8000 -s -w -c 1";

	if( (-e $in_middle_file) && (-e $out_middle_file) ) {
		$command = "soxmix -t $type '$in_middle_file' -t $type '$out_middle_file' $wav_options '$output_pathname'".".wav";
		run_cmd(0, $command) || return;
	} elsif (-e $in_middle_file) {
		$command = "sox -t $type '$in_middle_file' $wav_options '$output_pathname'".".wav";
		run_cmd(0, $command) || return;
	} elsif (-e $out_middle_file) {
		$command = "sox -t $type '$out_middle_file' $wav_options '$output_pathname'".".wav";
		run_cmd(0, $command) || return;
	}

	unlink $in_middle_file;
	unlink $out_middle_file;
	
	print "--- Done: ".$single_file." => ".$output_pathname.".wav"."\n";

	return $output_pathname.".wav";
}

sub do_batch_trans
{
	my @threads;
	my $thread_num;
	my $file_num;
	my $process_seconds;
	my $max_threads;
	my $estimate_seconds;
	my $start_time;
	my $end_time;

	# How much time expected to process this batch of files
	$process_seconds = $options{'options'}{'process_seconds'} ? $options{'options'}{'process_seconds'} : 60;

	# The maximum number of threads allowed to be created
	$max_threads = $options{'options'}{'max_threads'} ? $options{'options'}{'max_threads'} : 32;

	$file_total_size = 0;

	get_file_list();

	$file_num = $file_queue->pending();

	if($file_num == 0 || $file_total_size == 0) {
		addlog("No recordings found.\n");
		return 0;
	}

	addlog("\n\n===============================================\n");
	addlog("--- $file_num recordings found.\n");
	addlog("--- Total file size: ".int($file_total_size / 1000)." KBytes.\n");

	# Perl does an floating point division instead of integral division
	$estimate_seconds = ( $file_total_size / 1000000 ) * $transcoding_seconds_per_mbytes;

	$thread_num = ceil($estimate_seconds / $process_seconds);
	
	if($thread_num > $max_threads) {
		$thread_num = $max_threads;
	}

	addlog("--- Creating $thread_num threads transcoding recordings...\n");

	$start_time = time();

	for (my $count = 1; $count <= $thread_num; $count++) {
		my $t = threads->new(\&process_thread, $count);
		push(@threads,$t);
	}

	foreach (@threads) {
		my $t_count = $_->join();
		adddbg("--- Thread $t_count finished.\n");
	}

	$end_time = time();

	# Calculate how many seconds it takes per Mbytes per thread to do the transcoding.
	if($file_total_size / 1000000 >= 1) {
		$transcoding_seconds_per_mbytes = ceil( ($end_time - $start_time) / ($file_total_size / 1000000 / $thread_num) );
		$transcoding_seconds_per_mbytes = ($transcoding_seconds_per_mbytes < 1) ? 1 : $transcoding_seconds_per_mbytes;	
	}

	addlog("--- Finish transcoding $file_num recordings, ".int($file_total_size / 1000)." KBytes, took ".($end_time - $start_time)." seconds. Seconds/Mbytes/Thread: $transcoding_seconds_per_mbytes\n");	

	return $file_num;
}

sub get_file_list
{
	find(\&add_to_file_list, $options{directories}{input});
}

sub add_to_file_list
{
	my %file:shared;
	my $size;

	# don't deal with directory
	return if (-d || !-e);

	return if(/.txt/);
	return if(/~/);

	$size = -s $_;
	$file_total_size += $size;

	return if($_ !~ /.*\-in\..*$/);

	$file{'dir'} = $File::Find::dir;
	$file{'name'} = $_;
	$file{'path'} = $File::Find::name;
	
	$file_queue->enqueue(\%file);
}

sub process_thread
{
        my $t_count = shift;
        adddbg("====== Thread $t_count started =====\n");

	while(my $file = $file_queue->dequeue_nb()) {
		handle_file($t_count, $file);
	}
	
	adddbg("====== Thread $t_count ended =====\n");
	return $t_count;
}

sub t_log
{
	my $t = shift;
	my $m = shift;

	addlog("[$t] ".$m);
}

sub t_dbg
{
        my $t = shift;
        my $m = shift;

        adddbg("[$t] ".$m);
}

sub handle_file
{
	my $t_count = shift;
	my $file_info = shift;

	my $file_dir = ${$file_info}{'dir'};
	my $file_name = ${$file_info}{'name'};
	my $file_path = ${$file_info}{'path'};

	t_dbg($t_count, "Process file $file_path\n");

	# check the file format
	my $found = 0;
	my $ext;

	foreach my $format (@ast_format)
	{
		if($file_name =~ /\.\Q$format\E$/)
		{
                        $found = 1;
                        $ext = $format;
                        last;
		}
	}
	return unless($found);

	my $output_filename;
	my $in_or_out;
	if($file_name =~ /(.*)\-in\.\Q$ext\E$/)
	{
		$output_filename = $1;
		$in_or_out = 1;
	} elsif($file_name =~ /(.*)\-out\.\Q$ext\E$/) {
		$output_filename = $1;
		$in_or_out = 0;
	} else {
		t_log($t_count, "Unknown file: ".$file_path);
		return;
	}

	# get the output directory
	my $output_pathname;
	my $output_dir;
	foreach my $input_dir ($options{directories}{input})
	{
		if($file_dir =~ /^\Q$input_dir\E(.*)/)
		{
			$output_dir = $options{directories}{output}.'/'.$1;
			$output_pathname = $output_dir."/".$output_filename;
			last;
		}
	}	

	my $save_u = umask();
	umask(0);
	mkpath($output_dir, { verbose => 1, mode => 0777 }) unless(-e $output_dir);
	umask($save_u);

	# trancode all formats other than alaw & ulaw into ulaw
        my $in_orig_file;
        my $out_orig_file;
        my $command;
        if($in_or_out)
        {
		$in_orig_file = $file_path;
		$out_orig_file = $file_dir."/".$output_filename."-out.".$ext;
	} else {
		$out_orig_file = $file_path;
		$in_orig_file = $file_dir."/".$output_filename."-in.".$ext;
	}

	if($ext ne "ulaw" && $ext ne "alaw")
	{
		if(-e $in_orig_file)
		{
                        $command = "/usr/sbin/asterisk -rx \"file convert ".$in_orig_file." ".$output_pathname."-in.ulaw\"";
                        run_cmd($t_count, $command) || return;
		}

        	if(-e $out_orig_file)
		{
        	        $command = "/usr/sbin/asterisk -rx \"file convert ".$out_orig_file." ".$output_pathname."-out.ulaw\"";
			run_cmd($t_count, $command) || return;
		}

		$ext = "ulaw";
	} else {
		copy($in_orig_file, $output_pathname."-in.".$ext);
		copy($out_orig_file, $output_pathname."-out.".$ext)
	}
	
	# merge in & out into one and encode it as mp3 format
	my $in_middle_file = $output_pathname."-in.".$ext;
	my $out_middle_file = $output_pathname."-out.".$ext;
	my $type = $sox_format{$ext};
	my $rate = 44100;	#it has to be 44100 or above, otherwize red5 won't stream it.

	if( (-e $in_middle_file) && (-e $out_middle_file) ) {
		$command = "soxmix -t $type '$in_middle_file' -t $type '$out_middle_file' -t mp3 -r $rate '$output_pathname.mp3'";
	} elsif (-e $in_middle_file) {
		$command = "sox -t $type $in_middle_file -t mp3 -r $rate $output_pathname.mp3";
	} elsif (-e $out_middle_file) {
		$command = "sox -t $type $out_middle_file -t mp3 -r $rate $output_pathname.mp3";
	}

	run_cmd($t_count, $command) || return;

	unlink $in_middle_file;
	unlink $out_middle_file;

	if ($options{general}{keep_audio_files} != 1) {
		unlink $in_orig_file;
		unlink $out_orig_file;
	}
	
	# We don't need anymore because we use asterisk to do transcoding and use sox to do mixing and mp3 encoding.

	#$type = $ffmpeg_format{$ext};
	#$command = "ffmpeg -f $type -ar 8000 -ab 64k -i $output_pathname.$ext -ar 44100 -ab 64k $output_pathname.mp3";

	#print "--- Command: ".$command."\n";
	#system($command);
	#if (0 == $?) {
	#	if ($options{general}{keep_audio_files} != 1) {
	#		unlink("$output_pathname.$ext");
	#	}
	#} else {
	#	warn "$command failed: $!";
	#}

	t_dbg($t_count, "DONE: ".$file_path." => ".$output_pathname.".mp3"."\n");
}

sub run_cmd
{
	my $t_count = shift;
	my $command = shift;
        t_dbg($t_count, "COMMAND: ".$command."\n");
        system($command);
        unless (0 == $?) {
                t_log($t_count, "Failed to run $command: $!\n");
		return 0;
        }
	return 1;
}
